// Rosette.cpp
// Interfaces for the Rosette class and CreditsView class

// Rosettes, a screensaver module for BeOS
// Jon Pimble -- fatrat@fat-rat.com


#include "Rosette.h"



extern "C" _EXPORT BScreenSaver *instantiate_screen_saver( BMessage *msg, image_id id )
{
  return new Rosette( msg, id ); 
} 


// Constructor
Rosette::Rosette( BMessage *archive, image_id image ) 
          : BScreenSaver( archive, image ) 
{ 
  // Nothing here...
}



void Rosette::StartConfig(BView *view) 
{ 
  // Create and attach credits view
  // Note: We don't have to keep track of the object we create, because it will be
  //  destroyed for us when its parent view is destroyed
  CreditsView *credits;
  
  credits = new CreditsView(view->Bounds(), "Credits");

  view->AddChild(credits);
}



// Setup screensaver parameters before inital drawing
status_t Rosette::StartSaver( BView *view, bool preview ) 
{
  // Time between calls to Draw(), in microseconds (1000000us == 1s)
  SetTickSize(TICKS);

  // Must return B_OK for screensaver to start
  return B_OK;
}
       
       

// Main drawing function
void Rosette::Draw(BView *view, int32 frame)
{
  // Scratch variables
  BPoint points[4];
  int i, k;
  float tmp_angle;
  
  
  // In the first frame, blank the background and initialize the rosettes
  if(frame == 0)
  {  
    // Seed for random number generator
    srand(time(NULL));
  
    // Black out background
    view->SetHighColor(0, 0, 0);
    view->FillRect(view->Bounds());


    // Decide on how many rosettes we're going to use (1 to maximum)
    num_flowers = (rand() % (MAX_FLOWERS - 1)) + 1;
    
    
    // Initialize all rosettes
    for(i = 0; i < num_flowers; i++)
    {    
      // Pick center, within bounds of our view
      flowers[i].x_pos = rand() % (int) view->Bounds().Width();
      flowers[i].y_pos = rand() % (int) view->Bounds().Height();
    
      // Pick deltas for x, y movement
      // The factor calculated first is to scale the movement to make it
      //  look ok; this was initially tested on a 800x600 screen and
      //  tweaked until it looked good :)
      // We convert to int to keep the movement from getting jittery - fractional pixel
      //  movements can cause "jumping" instead of smooth movement
      flowers[i].dx = int((view->Bounds().Width() / 800.0) * ((rand() % 5) - (rand() % 5)));
      flowers[i].dy = int((view->Bounds().Height() / 600.0) * ((rand() % 5) - (rand() % 5)));
      
      // Make sure that we're moving at least a little in each direction
      if(flowers[i].dx == 0)
      {
        flowers[i].dx = 1.0;
      }
      if(flowers[i].dy == 0)
      {
        flowers[i].dy = 1.0;
      }
    

      // Pick initial offset angle
      flowers[i].petal_ofs_angle = rand() % 360;
    

      // Pick radii for circles
      // Center is at max 1/50 of screen size, minimum of 1
      flowers[i].d_r1 = fabs((rand() % (int) view->Bounds().Width())/50 + 1);

      // Middle radius is offset by r1, and is between 10 and 1/20 of screen size
      flowers[i].d_r2 = fabs((rand() % (int) view->Bounds().Width())/20 + 10);
    
      // Outer radius is offset by r1+r2, and is
      //  between 1 and 1/10 of screen size
      flowers[i].d_r3 = fabs((rand() % (int) view->Bounds().Width())/10 + 1);
    
    
      // Number of petals is 2 to MAX_PETALS
      flowers[i].num_petals = ((int) rand() % (MAX_PETALS - 2)) + 2;
    

      // Degree measure of offset for middle petal coordinates; this ranges from
      //  5 degrees to ((360/num_petals)/2), or 180/num_petals
      flowers[i].petal_mid_ofs_angle = (rand() % ((180/flowers[i].num_petals) - 5) ) + 5;
    

      // Pick a starting color
      // Hue is in degrees
      flowers[i].petal_color.h = flowers[i].petal_ofs_angle;
      // Lower saturation yeilds more subdued colors (more like pastels...)
      flowers[i].petal_color.s = (rand() % (int) 50)/100.0 + 0.3;
      // Keep overall brightness > 0.7 (just looks better)
      flowers[i].petal_color.v = (rand() % (int) 30)/100.0 + 0.7;
      // Alpha value is always 1.0 (full)
      flowers[i].petal_color.alpha = 1.0;
    }
  }
  

  // Used to sync to vertical retrace; this makes for a smoother display on some systems, as
  //  drawing will take place during the vertical blank -- and less flickering will be evident
  BScreen screen;
  screen.WaitForRetrace();
  
  
  // In every frame after the first, animate and draw all of the rosettes
  for(k = 0; k < num_flowers; k++)
  {
    // Center position of Rosette bounces around the screen
    flowers[k].x_pos += flowers[k].dx;
    flowers[k].y_pos += flowers[k].dy;
  
    // If we've gone off an edge, bounce
    if( ((flowers[k].x_pos < 0) && (flowers[k].dx < 0))
     || ((flowers[k].x_pos > view->Bounds().Width()) && (flowers[k].dx > 0)))
    {
      flowers[k].dx = flowers[k].dx * -1.0;
    }

    if( ((flowers[k].y_pos < 0) && (flowers[k].dy < 0))
     || ((flowers[k].y_pos > view->Bounds().Height()) && (flowers[k].dy > 0)))
    {
      flowers[k].dy = flowers[k].dy * -1.0;
    }
  
    // Update angle for middle points to expand from
    flowers[k].petal_mid_ofs_angle = fabs(sinf(flowers[k].petal_ofs_angle * DEG2RAD)) * 180/flowers[k].num_petals;

    // Make the rosette
    flowers[k].petal_ofs_angle = fmod(++flowers[k].petal_ofs_angle, 360);

    // Cycle colors
    flowers[k].petal_color.h = flowers[k].petal_ofs_angle;
    
    // Calculate current radii
    flowers[k].r1 = (fabs(sinf(flowers[k].petal_ofs_angle * DEG2RAD))) * flowers[k].d_r1 + 1;
    flowers[k].r2 = (fabs(sinf(flowers[k].petal_ofs_angle * DEG2RAD))) * flowers[k].d_r2 + 1;
    flowers[k].r3 = (fabs(sinf(flowers[k].petal_ofs_angle * DEG2RAD))) * flowers[k].d_r3 + 1;


    // Draw a rosette
    for(i = 1; i <= flowers[k].num_petals; i++)
    {

      tmp_angle = ((360.0/flowers[k].num_petals)*i  + flowers[k].petal_ofs_angle);
    
      // Start at a point on the circle defined by r1
      points[0].Set(
       flowers[k].x_pos + (cosf(tmp_angle * DEG2RAD) * flowers[k].r1),
       flowers[k].y_pos + (sinf(tmp_angle * DEG2RAD) * flowers[k].r1) );
    
      // Midpoints - one to the left and one to the right of the current angle,
      //  offset by the mid-circle offset angle; points are along the circle
      //  defined by r2
      points[1].Set(
       flowers[k].x_pos + cosf((tmp_angle - flowers[k].petal_mid_ofs_angle) * DEG2RAD) * (flowers[k].r1+flowers[k].r2),
       flowers[k].y_pos + sinf((tmp_angle - flowers[k].petal_mid_ofs_angle ) * DEG2RAD) * (flowers[k].r1+flowers[k].r2) );
    
      points[3].Set(
       flowers[k].x_pos + cosf((tmp_angle + flowers[k].petal_mid_ofs_angle) * DEG2RAD) * (flowers[k].r1+flowers[k].r2),
       flowers[k].y_pos + sinf((tmp_angle + flowers[k].petal_mid_ofs_angle) * DEG2RAD) * (flowers[k].r1+flowers[k].r2) );

    
      // Point on circle defined by r3
      points[2].Set(
       flowers[k].x_pos + (cosf(tmp_angle * DEG2RAD) * (flowers[k].r1+flowers[k].r2+flowers[k].r3)),
       flowers[k].y_pos + (sinf(tmp_angle * DEG2RAD) * (flowers[k].r1+flowers[k].r2+flowers[k].r3)) );
     
    
      // Fill in the petal, black
      view->SetHighColor(0, 0, 0);
      view->FillPolygon(points, 4);

      // Now plot the outline of the petal, with the current color
      view->SetHighColor(hsv_to_rgb(flowers[k].petal_color));
      view->StrokePolygon(points, 4);

    }
  }

}



// Convert an HSV color value to an RGB one
rgb_color Rosette::hsv_to_rgb(hsv_color hsv_to_convert)
{
  float f, p1, p2, p3, scaled_hue;
  float red, green, blue;
  rgb_color rgb_triplet;
  
  // Be sure that hue is in the range we want (0 to 359.9999)
  hsv_to_convert.h = fabs(fmodf(fabs(hsv_to_convert.h), 360));
  
  
  // Be sure that s, v are within acceptable range (0 to 1.0)
  if( (hsv_to_convert.s < 0) || (hsv_to_convert.s > 1.0) )
  {
    hsv_to_convert.s = fabs(fmodf(hsv_to_convert.s, 1.0));
  }
  if( (hsv_to_convert.v < 0) || (hsv_to_convert.v > 1.0) )
  {
    hsv_to_convert.v = fabs(fmodf(hsv_to_convert.v, 1.0));
  }
  
  
  // We want to deal with a hue value between 0 and 6 (corners on the HSV
  //  hexagon model)
  scaled_hue = hsv_to_convert.h / 60.0;
    
  // Scale hue to a value between 0 and 1.0
  f = scaled_hue - floor(scaled_hue);

  // Coordinate conversion from HSV hexagon to RGB cube
  p1 = hsv_to_convert.v * (1.0 - hsv_to_convert.s);
  p2 = hsv_to_convert.v * (1.0 - (hsv_to_convert.s * f));
  p3 = hsv_to_convert.v * (1.0 - (hsv_to_convert.s * (1.0 - f)));


  // Pick RGB values based on coordinates in HSV hexagon
  switch((int) floor(scaled_hue))
  {
    case 0:
      red = hsv_to_convert.v;
      green = p3;
      blue = p1;
      break;
        
    case 1:
      red = p2;
      green = hsv_to_convert.v;
      blue = p1;
      break;
        
    case 2:
      red = p1;
      green = hsv_to_convert.v;
      blue = p3;
      break;
        
    case 3:
      red = p1;
      green = p2;
      blue = hsv_to_convert.v;
      break;
        
     case 4:
      red = p3;
      green = p1;
      blue = hsv_to_convert.v;
      break;
        
    case 5:
      red = hsv_to_convert.v;
      green = p1;
      blue = p2;
      break;
        
    // If we made it to the default case, something bad happened!
    //  Set the rgb_color values to 0,0,0 as default
    default:
      red = 0;
      green = 0;
      blue = 0;
      break;
  }
  
  
  // Scale color values to 0, 255 for the color_rgb values
  rgb_triplet.red = (uint8) (red * 255.0);
  rgb_triplet.green = (uint8) (green * 255.0);
  rgb_triplet.blue = (uint8) (blue * 255.0);
  rgb_triplet.alpha = (uint8) (hsv_to_convert.alpha * 255.0);
  
  
  // Return the color_rgb triplet
  return rgb_triplet;
}




// Stuff for credits view

// Constructor
CreditsView::CreditsView(BRect frame, const char *name)
 : BView(frame, name, B_FOLLOW_ALL, B_WILL_DRAW)
{
  // Nothing to do here...
}


// Over-ride the virtual draw for the view, and display credits
void CreditsView::Draw(BRect r)
{
  // Dislpay some credits in the config view
  BFont font;
  font_height height;
  float y_pos, font_size, tmp;
  
  
  // Fill view with UI color
  this->SetLowColor(ui_color(B_PANEL_BACKGROUND_COLOR));
  this->FillRect(this->Bounds(), B_SOLID_LOW);


  // Get current font info
  this->GetFont(&font); 
  font.GetHeight(&height);


  // Remember what the current font size is...
  font_size = font.Size();
  
  // Figure out where to place this text - need height info
  //  ( * 2 here because two lines will be done in this size)
  tmp = 2 * (height.ascent + height.descent);
  
  // Go to a larger font size
  font.SetSize(font_size + 10);  

  // Figure out where to place this text - need height for new font size
  font.GetHeight(&height);
  
  // Set starting row position
  y_pos = (this->Bounds().Height() - tmp - height.ascent - height.descent)/2;
  
  
  // Draw the title, slightly larger font
  this->SetHighColor(0, 0, 0);
  this->SetFont(&font, B_FONT_ALL);
  this->MovePenTo(
   ((this->Bounds().right / 2) - (this->StringWidth(CREDITS_SSNAME) / 2) ), y_pos);
  
  this->DrawString(CREDITS_SSNAME, strlen(CREDITS_SSNAME), NULL);
  y_pos += (height.ascent + height.descent);
  
  
  // Go back to standard size font
  font.SetSize(font_size); 
  this->SetFont(&font, B_FONT_ALL);
  font.GetHeight(&height);

  // Version and date
  this->MovePenTo(
   ((this->Bounds().right / 2) - (this->StringWidth(CREDITS_VERSION) / 2) ), y_pos);
  this->DrawString(CREDITS_VERSION, strlen(CREDITS_VERSION), NULL);
  
  // Leave some extra space
  y_pos += height.ascent + height.descent;
  
  // Credit line
  this->MovePenTo(
   ((this->Bounds().right / 2) - (this->StringWidth(CREDITS_AUTHOR) / 2) ), y_pos);
  this->DrawString(CREDITS_AUTHOR, strlen(CREDITS_AUTHOR), NULL);
}

